from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from numpy.testing import assert_array_equal, assert_allclose
import kdcount.correlate as correlate
import os
import pytest

# debug logging
setup_logging("debug")

def gather_data(source, name):
    return numpy.concatenate(source.comm.allgather(source[name].compute()), axis=0)

def kcount_paircount(pos1, w1, redges, Nmu, boxsize, pos2=None, w2=None, los=2):
    """
    Verify the paircount algorithm using kdcount
    """
    # get the trees
    tree1 = correlate.points(pos1, boxsize=boxsize, weights=w1)
    if pos2 is None:
        tree2 = tree1
    else:
        tree2 = correlate.points(pos2, boxsize=boxsize, weights=w2)

    # setup the binning
    if Nmu == 0:
        bins = correlate.RBinning(redges)
    else:
        bins = correlate.FlatSkyBinning(redges, Nmu, los=los, mu_min=0., absmu=True,)

    # do the paircount
    pc = correlate.paircount(tree1, tree2, bins, np=0, usefast=False, compute_mean_coords=True)

    ravg = pc.mean_centers if Nmu == 0 else pc.mean_centers[0]
    return numpy.nan_to_num(pc.pair_counts), numpy.nan_to_num(ravg), pc.sum1

@MPITest([1, 3])
def test_1d_periodic_auto(comm):

    CurrentMPIComm.set(comm)

    # uniform source of particles
    BoxSize = 512.
    source = UniformCatalog(nbar=3e-6, BoxSize=BoxSize, seed=42)

    # add some weights b/w 0 and 1
    source['Weight'] = source.rng.uniform(size=len(source))

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # do the weighted paircount
    r = SimulationBoxPairCount('1d', source, redges, periodic=True, weight='Weight')

    # cannot compute r=0
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('1d', source, numpy.linspace(0, 10.0, 10))

    pos = gather_data(source, "Position")
    w = gather_data(source, "Weight")

    # verify with kdcount
    npairs, ravg, wsum = kcount_paircount(pos, w, redges, 0, source.attrs['BoxSize'])
    assert_allclose(ravg, r.result['r'])
    assert_allclose(npairs, r.result['npairs'])
    assert_allclose(wsum, r.result['npairs'] * r.result['weightavg'])

    # test save
    r.save('paircount-test.json')
    r2 = SimulationBoxPairCount.load('paircount-test.json')
    assert_array_equal(r.result.data, r2.result.data)

    if comm.rank == 0: os.remove('paircount-test.json')

@MPITest([1, 3])
def test_1d_nonperiodic_auto(comm):
    CurrentMPIComm.set(comm)

    # uniform source of particles
    BoxSize = 512.
    source = UniformCatalog(nbar=3e-6, BoxSize=BoxSize, seed=42)
    source['Weight'] = numpy.random.random(size=len(source))

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # do the weighted paircount
    r = SimulationBoxPairCount('1d', source, redges, periodic=False, weight='Weight')

    pos = gather_data(source, "Position")
    w = gather_data(source, "Weight")

    # verify with kdcount
    npairs, ravg, wsum = kcount_paircount(pos, w, redges, 0, None)
    assert_allclose(ravg, r.result['r'])
    assert_allclose(npairs, r.result['npairs'])
    assert_allclose(wsum, r.result['npairs'] * r.result['weightavg'])


@MPITest([1, 3])
def test_1d_periodic_cross(comm):

    CurrentMPIComm.set(comm)

    BoxSize = 512.
    first = UniformCatalog(nbar=3e-6, BoxSize=BoxSize, seed=42)
    second = UniformCatalog(nbar=3e-6, BoxSize=BoxSize, seed=84)

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # do the paircount
    r = SimulationBoxPairCount('1d', first, redges, second=second, periodic=True)

    pos1 = gather_data(first, "Position")
    pos2 = gather_data(second, "Position")

    # verify with kdcount
    npairs, ravg, wsum = kcount_paircount(pos1, None, redges, 0, first.attrs['BoxSize'], pos2=pos2)
    assert_allclose(ravg, r.result['r'])
    assert_allclose(npairs, r.result['npairs'])
    assert_allclose(wsum, r.result['npairs'] * r.result['weightavg'])

@MPITest([1, 3])
def test_2d_periodic_auto(comm):

    CurrentMPIComm.set(comm)

    # uniform source of particles
    BoxSize = 512.
    source = UniformCatalog(nbar=3e-6, BoxSize=BoxSize, seed=42)

    # add some weights b/w 0 and 1
    source['Weight'] = source.rng.uniform(size=len(source))

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)
    Nmu = 10

    # do the weighted paircount
    r = SimulationBoxPairCount('2d', source, redges, periodic=True, weight='Weight', Nmu=10)

    pos = gather_data(source, "Position")
    w = gather_data(source, "Weight")

    # verify with kdcount
    npairs, ravg, wsum = kcount_paircount(pos, w, redges, Nmu, source.attrs['BoxSize'])
    assert_allclose(ravg, r.result['r'])
    assert_allclose(npairs, r.result['npairs'])
    assert_allclose(wsum, r.result['npairs'] * r.result['weightavg'])

@MPITest([3])
def test_2d_diff_los(comm):

    CurrentMPIComm.set(comm)

    # uniform source of particles
    BoxSize = 512.
    source = UniformCatalog(nbar=3e-6, BoxSize=BoxSize, seed=42)

    # add some weights b/w 0 and 1
    source['Weight'] = source.rng.uniform(size=len(source))

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)
    Nmu = 10

    # do the weighted paircount
    r = SimulationBoxPairCount('2d', source, redges, periodic=True, weight='Weight', Nmu=10, los='x')

    pos = gather_data(source, "Position")
    w = gather_data(source, "Weight")

    # verify with kdcount
    npairs, ravg, wsum = kcount_paircount(pos, w, redges, Nmu, source.attrs['BoxSize'], los=0)
    assert_allclose(ravg, r.result['r'])
    assert_allclose(npairs, r.result['npairs'])
    assert_allclose(wsum, r.result['npairs'] * r.result['weightavg'])

@MPITest([1, 3])
def test_2d_nonperiodic_auto(comm):
    CurrentMPIComm.set(comm)

    # uniform source of particles
    source = UniformCatalog(nbar=3e-6, BoxSize=512, seed=42)
    source['Weight'] = numpy.random.random(size=len(source))

    # make the bin edges
    redges = numpy.linspace(10, 40., 10)
    Nmu = 10

    # do the weighted paircount
    r = SimulationBoxPairCount('2d', source, redges, periodic=False, weight='Weight', Nmu=10)

    pos = gather_data(source, "Position")
    w = gather_data(source, "Weight")

    # verify with kdcount
    npairs, ravg, wsum = kcount_paircount(pos, w, redges, Nmu, None)
    assert_allclose(ravg, r.result['r'])
    assert_allclose(npairs, r.result['npairs'])
    assert_allclose(wsum, r.result['npairs'] * r.result['weightavg'])


@MPITest([1, 3])
def test_2d_periodic_cross(comm):

    CurrentMPIComm.set(comm)

    BoxSize = 512.
    first = UniformCatalog(nbar=3e-6, BoxSize=BoxSize, seed=42)
    second = UniformCatalog(nbar=3e-6, BoxSize=BoxSize, seed=84)

    # make the bin edges
    redges = numpy.linspace(10, 40., 10)
    Nmu = 10

    # do the paircount
    r = SimulationBoxPairCount('2d', first, redges, second=second, periodic=True, Nmu=10)

    pos1 = gather_data(first, "Position")
    pos2 = gather_data(second, "Position")

    # verify with kdcount
    npairs, ravg, wsum = kcount_paircount(pos1, None, redges, Nmu, first.attrs['BoxSize'], pos2=pos2)
    assert_allclose(ravg, r.result['r'])
    assert_allclose(npairs, r.result['npairs'])
    assert_allclose(wsum, r.result['npairs'] * r.result['weightavg'])

@MPITest([1])
def test_bad_los(comm):

    CurrentMPIComm.set(comm)

    # uniform source of particles
    source = UniformCatalog(nbar=3e-6, BoxSize=512., seed=42)

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # should be 'x', 'y', 'z'
    with pytest.raises(AssertionError):
        r = SimulationBoxPairCount('1d', source, redges, los='a')

    # should be [0,1,2]
    with pytest.raises(AssertionError):
        r = SimulationBoxPairCount('1d', source, redges, los=3)

    # negative okay
    r = SimulationBoxPairCount('1d', source, redges, los=-1)

    # vector is bad
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('1d', source, redges, los=[0,0,1])

@MPITest([1])
def test_bad_rmax(comm):

    CurrentMPIComm.set(comm)

    # uniform source of particles
    source = UniformCatalog(nbar=3e-6, BoxSize=512., seed=42)

    # make the bin edges
    redges = numpy.linspace(10, 400., 10)

    # rmax is too big
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('1d', source, redges, periodic=True)

@MPITest([1])
def test_noncubic_box(comm):

    CurrentMPIComm.set(comm)

    # uniform source of particles
    source = UniformCatalog(nbar=3e-6, BoxSize=[512., 512., 256], seed=42)

    # make the bin edges
    redges = numpy.linspace(10, 100., 10)

    # cannot do non-cubic boxes
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('1d', source, redges, periodic=True)

@MPITest([1])
def test_bad_boxsize(comm):

    CurrentMPIComm.set(comm)

    # uniform source of particles
    first = UniformCatalog(nbar=3e-5, BoxSize=512.0, seed=42)
    second = UniformCatalog(nbar=3e-5, BoxSize=256.0, seed=42)

    # make the bin edges
    redges = numpy.linspace(10, 50., 10)

    # box sizes are different
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('1d', first, redges, second=second)

    # specified different value
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('1d', second, redges, BoxSize=300.)

    # no boxSize
    del first.attrs['BoxSize']
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('1d', first, redges)



from .sim_paircount import PairCountBase, verify_input_sources, MissingCorrfuncError
from nbodykit.binned_statistic import BinnedStatistic

import logging
import numpy

class SurveyPairCountBase(PairCountBase):
    """
    An abstract base class for pair count algorithms from surveys. The input
    data is assumed to be of the form (ra, dec, redshift) for these algorithms.
    """
    def _decompose(self, rmax, angular=False):
        """
        An internal function to perform a domain decomposition on the two
        sources we wish to correlate, returning the position/weights for each
        object in the correlating pair.

        The domain decomposition is based on the Cartesian coordinates of
        the input data (assumed to be in sky coordinates).

        The implementation follows:

        1. Decompose the first source and balance the particle load, such that
           the first source is evenly distributed across all ranks and the
           objects are spatially tight on a given rank.
        2. Decompose the second source, ensuring a given rank holds all
           particles within the desired maximum separation.

        Parameters
        ----------
        rmax : float
            the maximum Cartesian separation implied by the user's binning
        angular : bool, optional
            if ``True``, the Cartesian positions used in the domain
            decomposition are on the unit sphere

        Returns
        -------
        (pos1, w1), (pos2, w2) : array_like
            the (decomposed) set of positions and weights to correlate
        """
        from nbodykit.transform import StackColumns
        from pmesh.domain import GridND
        from nbodykit.utils import split_size_3d

        # either (ra,dec) or (ra,dec,redshift)
        poscols = [self.attrs['ra'], self.attrs['dec']]
        if not angular: poscols += [self.attrs['redshift']]

        # determine processor division for domain decomposition
        np = split_size_3d(self.comm.size)
        if self.comm.rank == 0:
            self.logger.info("using cpu grid decomposition: %s" %str(np))

        # stack position and compute
        pos1 = StackColumns(*[self.first[col] for col in poscols])
        pos1, w1 = self.first.compute(pos1, self.first[self.attrs['weight']])
        N1 = self.comm.allreduce(len(pos1))

        # get comoving dist and boxsize for first
        cosmo = self.attrs.get('cosmo', None)
        cpos1, boxsize1, rdist1 = get_cartesian(self.comm, pos1, cosmo=cosmo)

        # pass in comoving dist to Corrfunc instead of redshift
        if not angular:
            pos1[:,2] = rdist1

        # set up position for second too
        if self.second is not None:

            # stack position and compute for "second"
            pos2 = StackColumns(*[self.second[col] for col in poscols])
            pos2, w2 = self.second.compute(pos2, self.second[self.attrs['weight']])
            N2 = self.comm.allreduce(len(pos2))

            # get comoving dist and boxsize
            cpos2, boxsize2, rdist2 = get_cartesian(self.comm, pos2, cosmo=cosmo)

            # pass in comoving distance instead of redshift
            if not angular:
                pos2[:,2] = rdist2
        else:
            pos2 = pos1
            w2 = w1
            N2 = N1
            boxsize2 = boxsize1
            cpos2 = cpos1

        # determine global boxsize
        if self.second is None:
            boxsize = boxsize1
        else:
            boxsizes = numpy.vstack([boxsize1, boxsize2])
            argmax = numpy.argmax(boxsizes, axis=0)
            boxsize = boxsizes[argmax, [0,1,2]]

        # initialize the domain
        # NOTE: over-decompose by factor of 2 to trigger load balancing
        grid = [
            numpy.linspace(0, boxsize[0], 2*np[0] + 1, endpoint=True),
            numpy.linspace(0, boxsize[1], 2*np[1] + 1, endpoint=True),
            numpy.linspace(0, boxsize[2], 2*np[2] + 1, endpoint=True),
        ]
        domain = GridND(grid, comm=self.comm)

        # balance the load
        domain.loadbalance(domain.load(cpos1))

        # decompose based on cartesian positions
        layout = domain.decompose(cpos1, smoothing=0)
        pos1   = layout.exchange(pos1)
        w1     = layout.exchange(w1)

        # get the position/weight of the secondaries
        if rmax > boxsize.max() * 0.25:
            pos2 = numpy.concatenate(self.comm.allgather(pos2), axis=0)
            w2   = numpy.concatenate(self.comm.allgather(w2), axis=0)
        else:
            layout  = domain.decompose(cpos2, smoothing=rmax)
            pos2 = layout.exchange(pos2)
            w2   = layout.exchange(w2)

        # log the sizes of the trees
        self._log(N1, N2, pos1, pos2)

        return (pos1, w1), (pos2, w2)

class SurveyDataPairCount(SurveyPairCountBase):
    """
    Count (weighted) pairs of objects from a survey data catalog using the
    :mod:`Corrfunc` package.

    This uses the:func:`Corrfunc.mocks.DDsmu_mocks.DDsmu_mocks`
    function to count pairs.

    Results are computed when the object is inititalized. See the documenation
    of :func:`~SurveyDataPairCount.run` for the attributes storing the
    results.

    .. note::

        The algorithm expects the positions of particles from a survey catalog
        be the sky coordinates, right ascension and declination, and redshift.
        To compute pair counts in a simulation box, using the Cartesian
        coordinate vectors, see :class:`SimulationBoxPairCount`.

    .. warning::
        The right ascension and declination columns should be specified
        in degrees.

    Parameters
    ----------
    mode : {'1d', '2d'}
        compute paircounts as a function of ``r`` and ``mu`` or just ``r``
    first : CatalogSource
        the first source of particles, providing the 'Position' column
    redges : array_like
        the radius bin edges; length of nbins+1
    cosmo : :class:`~nbodykit.cosmology.core.Cosmology`
        the cosmology instance used to convert redshift into comoving distance
    second : CatalogSource, optional
        the second source of particles to cross-correlate
    Nmu : int, optional
        the number of ``mu`` bins, ranging from 0 to 1
    ra : str, optional
        the name of the column in the source specifying the
        right ascension coordinates in units of degrees; default is 'RA'
    dec : str, optional
        the name of the column in the source specifying the declination
        coordinates; default is 'DEC'
    redshift : str, optional
        the name of the column in the source specifying the redshift
        coordinates; default is 'Redshift'
    weight : str, optional
        the name of the column in the source specifying the object weights
    show_progress : bool, optional
        if ``True``, perform the pair counting calculation in 10 iterations,
        logging the progress after each iteration; this is useful for
        understanding the scaling of the code
    **config : key/value pairs
        additional keywords to pass to the :mod:`Corrfunc` function
    """
    logger = logging.getLogger('SurveyDataPairCount')

    def __init__(self, mode, first, redges, cosmo,
                    second=None, Nmu=5, ra='RA', dec='DEC',
                    redshift='Redshift', weight='Weight', show_progress=True,
                    **config):

        # check input 'mode'
        assert mode in ['1d', '2d'], "PairCount mode must be '1d' or '2d'"

        # check rmin
        if numpy.min(redges) <= 0.:
            raise ValueError(("the lower edge of the 1st separation bin must "
                              "greater than zero (no self-pairs)"))

        # verify the input sources
        required_cols = [ra, dec, redshift, weight]
        verify_input_sources(first, second, None, required_cols, inspect_boxsize=False)

        # init the base class
        SurveyPairCountBase.__init__(self, first, second, show_progress)

        # save the meta-data
        self.attrs['mode'] = mode
        self.attrs['redges'] = redges
        self.attrs['Nmu'] = Nmu
        self.attrs['cosmo'] = cosmo
        self.attrs['weight'] = weight
        self.attrs['ra'] = ra
        self.attrs['dec'] = dec
        self.attrs['redshift'] = redshift
        self.attrs['config'] = config

        # run the algorithm
        self.run()

    def run(self):
        """
        Calculate the 3D pair-counts of a survey data catalog as a function
        of separation ``r`` or separation and angle to line-of-sight
        (``r``, ``mu``). This adds the following attribute:

        - :attr:`SurveyDataPairCount.result`

        Attributes
        ----------
        result : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            a BinnedStatistic object holding the pair count results. The
            coordinate grid is either ``r`` or ``r`` and ``mu``.
            It stores the following variables:

            - ``r``: the mean separation value in the bin
            - ``npairs``: the number of pairs in the bin
            - ``weightavg``: the average weight value in the bin; each pair
              contributes the product of the individual weight values
        """
        try:
            from Corrfunc.mocks import DDsmu_mocks
        except ImportError:
            raise MissingCorrfuncError()

        # some setup
        redges = self.attrs['redges']
        Nmu = 1 if self.attrs['mode'] == '1d' else self.attrs['Nmu']

        # maximum separation value
        rmax = numpy.max(self.attrs['redges'])

        # domain decompose the data
        # NOTE: pos has 3 columns: (ra, dec, redshift)
        (pos1, w1), (pos2, w2) = self._decompose(rmax, angular=False)

        # do the pair counting
        kws = {}
        kws['autocorr'] = 0
        kws['cosmology'] = 1
        kws['nthreads'] = 1
        kws['nmu_bins'] = Nmu
        kws['mu_max'] = 1.0
        kws['binfile'] = self.attrs['redges']
        kws['RA2'] = pos2[:,0]
        kws['DEC2'] = pos2[:,1]
        kws['CZ2'] = pos2[:,2]
        kws['weights2'] = w2.astype(pos2.dtype)
        kws['weight_type'] = 'pair_product'
        kws['output_savg'] = True
        kws['is_comoving_dist'] = True
        kws.update(self.attrs['config'])

        def callback(kws, chunk):
            kws['RA1'] = pos1[chunk][:,0] # ra
            kws['DEC1'] = pos1[chunk][:,1] # dec
            kws['CZ1'] = pos1[chunk][:,2] # comoving distance
            kws['weights1'] = w1[chunk].astype(pos1.dtype)

        # run
        sizes = self.comm.allgather(len(pos1))
        pc = self._run(DDsmu_mocks, kws, sizes, callback=callback)
        pc = pc.reshape((-1, Nmu))

        # sum results over all ranks
        pc = self.comm.allreduce(pc)

        # make a new structured array
        dtype = numpy.dtype([('r', 'f8'), ('npairs', 'u8'), ('weightavg', 'f8')])
        data = numpy.zeros(pc.shape, dtype=dtype)

        # copy over main results
        data['r'] = pc['savg']
        data['npairs'] = pc['npairs']
        data['weightavg'] = pc['weightavg']

        # make the BinnedStatistic
        if self.attrs['mode'] == '1d':
            self.result = BinnedStatistic(['r'], [redges], numpy.squeeze(data), fields_to_sum=['npairs'])
        else:
            muedges = numpy.linspace(0, 1., Nmu+1)
            self.result = BinnedStatistic(['r','mu'], [redges,muedges], data, fields_to_sum=['npairs'])

class AngularPairCount(SurveyPairCountBase):
    """
    Count (weighted) angular pairs of objects from a survey data catalog using the
    :mod:`Corrfunc` package.

    This uses the:func:`Corrfunc.mocks.DDtheta_mocks.DDtheta_mocks`
    function to count pairs.

    Results are computed when the object is inititalized. See the documenation
    of :func:`~AngularPairCount.run` for the attributes storing the
    results.

    .. note::

        The algorithm expects the positions of particles from the survey catalog
        be the right ascension and declination angular coordinates.

    .. warning::
        The right ascension and declination columns should be specified
        in degrees.

    Parameters
    ----------
    first : CatalogSource
        the first source of particles, providing the 'Position' column
    edges : array_like
        the angular separation bin edges (in degrees); length of ``nbins+1``
    second : CatalogSource, optional
        the second source of particles to cross-correlate
    ra : str, optional
        the name of the column in the source specifying the
        right ascension coordinates in units of degrees; default is 'RA'
    dec : str, optional
        the name of the column in the source specifying the declination
        coordinates; default is 'DEC'
    weight : str, optional
        the name of the column in the source specifying the object weights
    show_progress : bool, optional
        if ``True``, perform the pair counting calculation in 10 iterations,
        logging the progress after each iteration; this is useful for
        understanding the scaling of the code
    **config : key/value pairs
        additional keywords to pass to the :mod:`Corrfunc` function
    """
    logger = logging.getLogger('AngularPairCount')

    def __init__(self, first, edges, second=None, ra='RA', dec='DEC',
                    weight='Weight', show_progress=True, **config):

        # check theta min
        if numpy.min(edges) <= 0.:
            raise ValueError(("the lower edge of the 1st separation bin must "
                              "greater than zero (no self-pairs)"))

        # verify the input sources
        verify_input_sources(first, second, None, [ra, dec, weight], inspect_boxsize=False)

        # init the base class
        SurveyPairCountBase.__init__(self, first, second, show_progress)

        # save the meta-data
        self.attrs['edges'] = edges
        self.attrs['ra'] = ra
        self.attrs['dec'] = dec
        self.attrs['weight'] = weight
        self.attrs['config'] = config

        # run the algorithm
        self.run()

    def __setstate__(self, state):
        self.__dict__.update(state)
        edges = self.attrs['edges']
        self.result = BinnedStatistic(['theta'], [edges], self.result, fields_to_sum=['npairs'])

    def run(self):
        """
        Calculate the angular pair counts of a survey data catalog as a function
        of separation angle on the sky. This adds the following attribute:

        - :attr:`AngularPairCount.result`

        Attributes
        ----------
        result : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            a BinnedStatistic object holding the pair count results.
            The coordinate grid is ``theta``, the angular separation bins.
            It stores the following variables:

            - ``theta``: the mean separation value in the bin
            - ``npairs``: the number of pairs in the bin
            - ``weightavg``: the average weight value in the bin; each pair
              contributes the product of the individual weight values
        """
        try:
            from Corrfunc.mocks import DDtheta_mocks
        except ImportError:
            raise MissingCorrfuncError()

        # maximum R separation from angular separation
        theta_max = numpy.max(self.attrs['edges'])
        rmax = 2 * numpy.sin(0.5 * numpy.deg2rad(theta_max))

        # domain decompose the data
        # NOTE: pos is angular and only has 2 columns: (ra, dec)
        (pos1, w1), (pos2, w2) = self._decompose(rmax, angular=True)

        # do the pair counting
        kws = {}
        kws['autocorr'] = 0
        kws['nthreads'] = 1
        kws['binfile'] = self.attrs['edges']
        kws['RA2'] = pos2[:,0]
        kws['DEC2'] = pos2[:,1]
        kws['weights2'] = w2.astype(pos2.dtype)
        kws['weight_type'] = 'pair_product'
        kws['output_thetaavg'] = True
        kws.update(self.attrs['config'])

        def callback(kws, chunk):
            kws['RA1'] = pos1[chunk][:,0] # ra
            kws['DEC1'] = pos1[chunk][:,1] # dec
            kws['weights1'] = w1[chunk].astype(pos1.dtype)

        # run
        sizes = self.comm.allgather(len(pos1))
        pc = self._run(DDtheta_mocks, kws, sizes, callback=callback)

        # sum results over all ranks
        pc = self.comm.allreduce(pc)

        # make a new structured array
        dtype = numpy.dtype([('theta', 'f8'), ('npairs', 'u8'), ('weightavg', 'f8')])
        data = numpy.zeros(pc.shape, dtype=dtype)

        # copy over main results
        data['theta'] = pc['thetaavg']
        data['npairs'] = pc['npairs']
        data['weightavg'] = pc['weightavg']

        # make the BinnedStatistic
        edges = [self.attrs['edges']]
        self.result = BinnedStatistic(['theta'], edges, data, fields_to_sum=['npairs'])

def get_cartesian(comm, pos, cosmo=None):
    """
    Convert sky coordinates to Cartesian coordinates and return the implied
    box size from the position bounds.

    If ``cosmo`` is not provided, return coordinates on the unit sphere.
    """
    from nbodykit.utils import get_data_bounds

    # get RA,DEC in degrees
    ra, dec = numpy.deg2rad(pos[:,0]), numpy.deg2rad(pos[:,1])

    # cartesian position
    x = numpy.cos( dec ) * numpy.cos( ra )
    y = numpy.cos( dec ) * numpy.sin( ra )
    z = numpy.sin( dec )
    cpos = numpy.vstack([x,y,z]).T

    # multiply by comoving distance?
    if cosmo is not None:
        assert pos.shape[-1] == 3
        rdist = cosmo.comoving_distance(pos[:,2]) # in Mpc/h
        cpos = rdist[:,None] * cpos
    else:
        rdist = None

    # min/max of position
    cpos_min, cpos_max = get_data_bounds(cpos, comm)
    boxsize = numpy.ceil(abs(cpos_max - cpos_min))

    return cpos, boxsize, rdist



from nbodykit.base.catalog import column
from nbodykit.source.catalog.array import ArrayCatalog
from nbodykit import CurrentMPIComm
from nbodykit.utils import GatherArray, ScatterArray

from six import add_metaclass
import abc
import logging
import numpy

def gal_type_integers(galtab):
    """
    Convert ``centrals`` to 0 and ``satellites`` to 1
    in the input galaxy Table

    This operatio is done in place
    """
    gal_type = numpy.zeros(len(galtab), dtype='i4')
    sats = galtab['gal_type'] == 'satellites'
    gal_type[sats] = 1
    galtab.replace_column('gal_type', gal_type)

def find_object_dtypes(data):
    """
    Utility function to convert 'O' data types to strings
    """
    for col in data.colnames:
        if data.dtype[col] == 'O':
            raise TypeError("column '%s' is of type 'O'; must convert to integer or string" %col)
    return data

class HODBase(ArrayCatalog):
    """
    A base class to be used for HOD population of a halo catalog.

    The user must supply the :func:`__makemodel__` function, which returns
    the halotools composite HOD model.

    This abstraction allows the user to potentially implement several
    different types of HOD models quickly, while using the population
    framework of this base class.
    """
    logger = logging.getLogger("HODBase")

    @CurrentMPIComm.enable
    def __init__(self, halos, seed=None, use_cache=False, comm=None, **params):

        from halotools.empirical_models import model_defaults

        # check type
        from halotools.sim_manager import UserSuppliedHaloCatalog
        if not isinstance(halos, UserSuppliedHaloCatalog):
            raise TypeError("input halos catalog for HOD should be halotools.sim_manager.UserSuppliedHaloCatalog")

        # try to extract meta-data from catalog
        mdef     = getattr(halos, 'mdef', 'vir')
        cosmo    = getattr(halos, 'cosmology', None)
        Lbox     = getattr(halos, 'Lbox', None)
        redshift = getattr(halos, 'redshift', None)

        # fail if we are missing any of these
        for name, attr in zip(['cosmology', 'Lbox', 'redshift'], [cosmo, Lbox, redshift]):
            if attr is None:
                raise AttributeError("input UserSuppliedHaloCatalog must have '%s attribute" %name)

        # promote astropy cosmology to nbodykit's Cosmology
        from astropy.cosmology import FLRW
        if isinstance(cosmo, FLRW):
            from nbodykit.cosmology import Cosmology
            cosmo = Cosmology.from_astropy(cosmo)

        # store the halotools catalog
        self._halos = halos
        self.cosmo  = cosmo
        self.comm   = comm

        # set the seed randomly if it is None
        if seed is None:
            if self.comm.rank == 0:
                seed = numpy.random.randint(0, 4294967295)
            seed = self.comm.bcast(seed)

        # grab the BoxSize from the halotools catalog
        self.attrs['BoxSize'] = numpy.empty(3)
        self.attrs['BoxSize'][:] = Lbox

        # store mass and radius keys
        self.mass   = model_defaults.get_halo_mass_key(mdef)
        self.radius = model_defaults.get_halo_boundary_key(mdef)

        # check for any missing columns
        needed = ['halo_%s' %s for s in ['x','y','z','vx','vy','vz', 'id', 'upid']] + [self.mass, self.radius]
        missing = set(needed) - set(halos.halo_table.colnames)
        if len(missing):
            raise ValueError("missing columns from halotools UserSuppliedHaloCatalog: %s" %str(missing))

        # this ensures that all columns in the halo catalog will propagate to the galaxy catalog
        model_defaults.default_haloprop_list_inherited_by_mock = halos.halo_table.colnames

        # store the attributes
        self.attrs['mdef'] = mdef
        self.attrs['redshift'] = redshift
        self.attrs['seed'] = seed
        self.attrs.update(params)

        # add cosmo attributes
        self.attrs['cosmo'] = dict(cosmo)

        # make the model!
        self._model = self.__makemodel__()

        # set the HOD params
        for param in self._model.param_dict:
            if param not in self.attrs:
                raise ValueError("missing '%s' parameter when initializing HOD" %param)
            self._model.param_dict[param] = self.attrs[param]

        # make the actual source
        ArrayCatalog.__init__(self, self.__makesource__(), comm=comm, use_cache=use_cache)

        # crash with no particles!
        if self.csize == 0:
            raise ValueError("no particles in catalog after populating HOD")

    @abc.abstractmethod
    def __makemodel__(self):
        """
        Abstract class to be overwritten by user; this should return
        the HOD model instance that will be used to do the mock
        population.

        See :ref:`the documentation <custom-hod-mocks>` for more details.

        Returns
        -------
        :class:`~halotools.empirical_models.HodModelFactory`
            the halotools object implementing the HOD model
        """
        pass

    def __makesource__(self):
        """
        Make the source of galaxies by performing the halo HOD population

        .. note::
            The mock population is only done by the root, and the resulting
            catalog is then distributed evenly amongst the available ranks
        """
        from astropy.table import Table

        # gather all halos to root
        halo_table = find_object_dtypes(self._halos.halo_table)
        all_halos = GatherArray(halo_table.as_array(), self.comm, root=0)

        # root does the mock population
        if self.comm.rank == 0:

            # set the halo table on the root to the Table containing all halo
            self._halos.halo_table = Table(data=all_halos, copy=True)
            del all_halos

            # populate
            self._model.populate_mock(halocat=self._halos, halo_mass_column_key=self.mass,
                                      Num_ptcl_requirement=1, seed=self.attrs['seed'])

            # remap gal_type to integers (cen: 0, sats: 1)
            gal_type_integers(self._model.mock.galaxy_table)

            # crash if any object dtypes
            data = find_object_dtypes(self._model.mock.galaxy_table).as_array()
            del self._model.mock.galaxy_table
        else:
            data = None

        # log the stats
        if self.comm.rank == 0:
            self._log_populated_stats(data)

        return ScatterArray(data, self.comm)

    def repopulate(self, seed=None, **params):
        """
        Update the HOD parameters and then re-populate the mock catalog

        .. warning::
            This operation is done in-place, so the size of the Source
            changes

        Parameters
        ----------
        seed : int, optional
            the new seed to use when populating the mock
        params :
            key/value pairs of HOD parameters to update
        """
        # set the seed randomly if it is None
        if seed is None:
            if self.comm.rank == 0:
                seed = numpy.random.randint(0, 4294967295)
            seed = self.comm.bcast(seed)
        self.attrs['seed'] = seed

        # update the HOD model parameters
        for name in params:
            if name not in self._model.param_dict:
                valid = list(self._model.param_dict.keys())
                raise ValueError("'%s' is not a valid Hod parameter name; valid are: %s" %(name, str(valid)))
            self._model.param_dict[name] = params[name]
            self.attrs[name] = params[name]

        # the root will do the mock population
        if self.comm.rank == 0:

            # re-populate the mock (without halo catalog pre-processing)
            self._model.mock.populate(Num_ptcl_requirement=1,
                                      halo_mass_column_key=self.mass,
                                      seed=self.attrs['seed'])

            # remap gal_type to integers (cen: 0, sats: 1)
            gal_type_integers(self._model.mock.galaxy_table)

            # crash if any object dtypes
            data = find_object_dtypes(self._model.mock.galaxy_table).as_array()
            del self._model.mock.galaxy_table

        else:
            data = None

        # log the stats
        if self.comm.rank == 0:
            self._log_populated_stats(data)

        # re-initialize with new source
        ArrayCatalog.__init__(self, ScatterArray(data, self.comm), comm=self.comm, use_cache=self.use_cache)

    def _log_populated_stats(self, data):
        """
        Internal function to log statistics of the populated catalog
        """
        if len(data) > 0:

            fsat = 1.*(data['gal_type'] == 1).sum()/len(data)
            self.logger.info("satellite fraction: %.2f" %fsat)
            self.attrs['fsat'] = fsat

            logmass = numpy.log10(data[self.mass])
            self.logger.info("mean log10 halo mass: %.2f" %logmass.mean())
            self.logger.info("std log10 halo mass: %.2f" %logmass.std())

    @column
    def Position(self):
        """
        Galaxy positions, in units of Mpc/h
        """
        pos = numpy.vstack([self._source['x'], self._source['y'], self._source['z']]).T
        return self.make_column(pos)

    @column
    def Velocity(self):
        """
        Galaxy velocity, in units of km/s
        """
        vel = numpy.vstack([self._source['vx'], self._source['vy'], self._source['vz']]).T
        return self.make_column(vel)

    @column
    def VelocityOffset(self):
        """
        The RSD velocity offset, in units of Mpc/h

        This multiplies Velocity by 1 / (a*100*E(z)) = 1 / (a H(z)/h)
        """
        z = self.attrs['redshift']
        rsd_factor = (1+z) / (100*self.cosmo.efunc(z))
        return self['Velocity'] * rsd_factor

class HODCatalog(HODBase):
    """
    A CatalogSource that uses the HOD prescription of Zheng et al 2007
    to populate an input halo catalog with galaxies.

    The mock population is done using :mod:`halotools`. See the documentation
    for :class:`halotools.empirical_models.Zheng07Cens` and
    :class:`halotools.empirical_models.Zheng07Sats` for further details
    regarding the HOD.

    The columns generated in this catalog are:

    #. **Position**: the galaxy position
    #. **Velocity**: the galaxy velocity
    #. **VelocityOffset**: the RSD velocity offset, in units of distance
    #. **conc_NFWmodel**: the concentration of the halo
    #. **gal_type**: the galaxy type, 0 for centrals and 1 for satellites
    #. **halo_id**: the global ID of the halo this galaxy belongs to, between 0 and :attr:`csize`
    #. **halo_local_id**: the local ID of the halo this galaxy belongs to, between 0 and :attr:`size`
    #. **halo_mvir**: the halo mass
    #. **halo_nfw_conc**: alias of ``conc_NFWmodel``
    #. **halo_num_centrals**: the number of centrals that this halo hosts, either 0 or 1
    #. **halo_num_satellites**: the number of satellites that this halo hosts
    #. **halo_rvir**: the halo radius
    #. **halo_upid**: equal to -1; should be ignored by the user
    #. **halo_vx, halo_vy, halo_vz**: the three components of the halo velocity
    #. **halo_x, halo_y, halo_z**: the three components of the halo position
    #. **host_centric_distance**: the distance from this galaxy to the center of the halo
    #. **vx, vy, vz**: the three components of the galaxy velocity, equal to ``Velocity``
    #. **x,y,z**: the three components of the galaxy position, equal to ``Position``

    For futher details, please see the :ref:`documentation <hod-mock-data>`.

    .. note::
         Default HOD values are from
         `Reid et al. 2014 <https://arxiv.org/abs/1404.3742>`_

    Parameters
    ----------
    halos : :class:`~halotools.sim_manager.UserSuppliedHaloCatalog`
        the halotools table holding the halo data; this object must have
        the following attributes: `cosmology`, `Lbox`, `redshift`
    logMmin : float, optional
        Minimum mass required for a halo to host a central galaxy
    sigma_logM : float, optional
        Rate of transition from <Ncen>=0 --> <Ncen>=1
    alpha : float, optional
        Power law slope of the relation between halo mass and <Nsat>
    logM0 : float, optional
        Low-mass cutoff in <Nsat>
    logM1 : float, optional
        Characteristic halo mass where <Nsat> begins to assume a power law form
    seed : int, optional
        the random seed to generate deterministic mocks

    References
    ----------
    `Zheng et al. (2007), arXiv:0703457 <https://arxiv.org/abs/astro-ph/0703457>`_
    """
    logger = logging.getLogger("HOD")

    @CurrentMPIComm.enable
    def __init__(self, halos, logMmin=13.031, sigma_logM=0.38,
                    alpha=0.76, logM0=13.27, logM1=14.08,
                    seed=None, use_cache=False, comm=None):

        params = {}
        params['logMmin'] = logMmin
        params['sigma_logM'] = sigma_logM
        params['alpha'] = alpha
        params['logM0'] = logM0
        params['logM1'] = logM1

        HODBase.__init__(self, halos, seed=seed, use_cache=use_cache, comm=comm, **params)

    def __repr__(self):
        names = ['logMmin', 'sigma_logM', 'alpha', 'logM0', 'logM1']
        s = ', '.join(['%s=%.2f' %(k,self.attrs[k]) for k in names])
        return "HODCatalog(%s)" %s

    def __makemodel__(self):
        """
        Return the Zheng 07 HOD model.

        This model evaluates Eqs. 2 and 5 of Zheng et al. 2007
        """
        from halotools.empirical_models import HodModelFactory
        from halotools.empirical_models import Zheng07Sats, Zheng07Cens
        from halotools.empirical_models import NFWPhaseSpace, TrivialPhaseSpace

        model = {}

        # use concentration from halo table
        if 'halo_nfw_conc' in self._halos.halo_table.colnames:
            conc_mass_model = 'direct_from_halo_catalog'
        # use empirical prescription for c(M)
        else:
            conc_mass_model = 'dutton_maccio14'

        # occupation functions
        cenocc = Zheng07Cens(prim_haloprop_key=self.mass)
        satocc = Zheng07Sats(prim_haloprop_key=self.mass, modulate_with_cenocc=True, cenocc_model=cenocc)
        satocc._suppress_repeated_param_warning = True

        # add to model
        model['centrals_occupation'] = cenocc
        model['satellites_occupation'] = satocc

        # profile functions
        kws = {'cosmology':self.cosmo.to_astropy(), 'redshift':self.attrs['redshift'], 'mdef':self.attrs['mdef']}
        model['centrals_profile'] = TrivialPhaseSpace(**kws)
        model['satellites_profile'] = NFWPhaseSpace(conc_mass_model=conc_mass_model, **kws)

        return HodModelFactory(**model)



import numpy
import dask.array as da
from six import string_types
from nbodykit.utils import deprecate

def StackColumns(*cols):
    """
    Stack the input dask arrays vertically, column by column.

    This uses :func:`dask.array.vstack`.

    Parameters
    ----------
    *cols : :class:`dask.array.Array`
        the dask arrays to stack vertically together

    Returns
    -------
    :class:`dask.array.Array` :
        the dask array where columns correspond to the input arrays

    Raises
    ------
    TypeError
        If the input columns are not dask arrays
    """
    if not all(isinstance(col, da.Array) for col in cols):
        raise TypeError("all input columns in `vstack` must be dask arrays")

    return da.vstack(cols).T

def ConcatenateSources(*sources, **kwargs):
    """
    Concatenate CatalogSource objects together, optionally including only
    certain columns in the returned source.

    .. note::
        The returned catalog object carries the meta-data from only
        the first catalog supplied to this function (in the ``attrs`` dict).

    Parameters
    ----------
    *sources : subclass of :class:`~nbodykit.base.catalog.CatalogSource`
        the catalog source objects to concatenate together
    columns : str, list of str, optional
        the columns to include in the concatenated catalog

    Returns
    -------
    CatalogSource :
        the concatenated catalog source object

    Examples
    --------
    >>> from nbodykit.lab import *
    >>> source1 = UniformCatalog(nbar=100, BoxSize=1.0)
    >>> source2 = UniformCatalog(nbar=100, BoxSize=1.0)
    >>> print(source1.csize, source2.csize)
    >>> combined = transform.ConcatenateSources(source1, source2, columns=['Position', 'Velocity'])
    >>> print(combined.csize)
    """
    from nbodykit.base.catalog import CatalogSource

    columns = kwargs.get('columns', None)
    if isinstance(columns, string_types):
        columns = [columns]

    # concatenate all columns, if none provided
    if columns is None or columns == []:
        columns = sources[0].columns

    # check comms
    if not all(src.comm == sources[0].comm for src in sources):
        raise ValueError("cannot concatenate sources: comm mismatch")

    # check all columns are there
    for source in sources:
        if not all(col in source for col in columns):
            raise ValueError(("cannot concatenate sources: columns are missing "
                              "from some sources"))
    # the total size
    size = sum(src.size for src in sources)

    data = {}
    for col in columns:
        data[col] = da.concatenate([src[col] for src in sources], axis=0)

    toret = CatalogSource._from_columns(size, sources[0].comm, use_cache=sources[0].use_cache, **data)
    toret.attrs.update(sources[0].attrs)
    return toret

def ConstantArray(value, size, chunks=100000):
    """
    Return a dask array of the specified ``size`` holding a single value.

    This uses numpy's "stride tricks" to avoid replicating
    the data in memory for each element of the array.

    Parameters
    ----------
    value : float
        the scalar value to fill the array with
    size : int
        the length of the returned dask array
    chunks : int, optional
        the size of the dask array chunks
    """
    ele = numpy.array(value)
    toret = numpy.lib.stride_tricks.as_strided(ele, [size] + list(ele.shape), [0] + list(ele.strides))
    return da.from_array(toret, chunks=chunks)


def SkyToUnitSphere(ra, dec, degrees=True):
    """
    Convert sky coordinates (``ra``, ``dec``) to Cartesian coordinates on
    the unit sphere.

    Parameters
    ----------
    ra : :class:`dask.array.Array`; shape: (N,)
        the right ascension angular coordinate
    dec : :class:`dask.array.Array`; ; shape: (N,)
        the declination angular coordinate
    degrees : bool, optional
        specifies whether ``ra`` and ``dec`` are in degrees or radians

    Returns
    -------
    pos : :class:`dask.array.Array`; shape: (N,3)
        the cartesian position coordinates, where columns represent
        ``x``, ``y``, and ``z``

    Raises
    ------
    TypeError
        If the input columns are not dask arrays
    """
    if not all(isinstance(col, da.Array) for col in [ra, dec]):
        raise TypeError("both ``ra`` and ``dec`` must be dask arrays")

    # put into radians from degrees
    if degrees:
        ra  = da.deg2rad(ra)
        dec = da.deg2rad(dec)

    # cartesian coordinates
    x = da.cos( dec ) * da.cos( ra )
    y = da.cos( dec ) * da.sin( ra )
    z = da.sin( dec )
    return da.vstack([x,y,z]).T

def SkyToCartesian(ra, dec, redshift, cosmo, degrees=True):
    """
    Convert sky coordinates (``ra``, ``dec``, ``redshift``) to a
    Cartesian ``Position`` column.

    .. warning::

        The returned Cartesian position is in units of Mpc/h.

    Parameters
    -----------
    ra : :class:`dask.array.Array`; shape: (N,)
        the right ascension angular coordinate
    dec : :class:`dask.array.Array`; shape: (N,)
        the declination angular coordinate
    redshift : :class:`dask.array.Array`; shape: (N,)
        the redshift coordinate
    cosmo : :class:`~nbodykit.cosmology.cosmology.Cosmology`
        the cosmology used to meausre the comoving distance from ``redshift``
    degrees : bool, optional
        specifies whether ``ra`` and ``dec`` are in degrees

    Returns
    -------
    pos : :class:`dask.array.Array`; shape: (N,3)
        the cartesian position coordinates, where columns represent
        ``x``, ``y``, and ``z`` in units of Mpc/h

    Raises
    ------
    TypeError
        If the input columns are not dask arrays
    """
    if not all(isinstance(col, da.Array) for col in [ra, dec, redshift]):
        raise TypeError("input ra, dec, and redshift objects must be dask arrays")

    # pos on the unit sphere
    pos = SkyToUnitSphere(ra, dec, degrees=degrees)

    # multiply by the comoving distance in Mpc/h
    r = redshift.map_blocks(lambda z: cosmo.comoving_distance(z), dtype=redshift.dtype)

    return r[:,None] * pos

def HaloConcentration(mass, cosmo, redshift, mdef='vir'):
    """
    Return halo concentration from halo mass, based on the analytic fitting
    formulas presented in
    `Dutton and Maccio 2014 <https://arxiv.org/abs/1402.7073>`_.

    .. note::
        The units of the input mass are assumed to be :math:`M_{\odot}/h`

    Parameters
    ----------
    mass : array_like
        either a numpy or dask array specifying the halo mass; units
        assumed to be :math:`M_{\odot}/h`
    cosmo : :class:`~nbodykit.cosmology.cosmology.Cosmology`
        the cosmology instance used in the analytic formula
    redshift : float
        compute the c(M) relation at this redshift
    mdef : str, optional
        string specifying the halo mass definition to use; should be
        'vir' or 'XXXc' or 'XXXm' where 'XXX' is an int specifying the
        overdensity

    Returns
    -------
    concen : :class:`dask.array.Array`
        a dask array holding the analytic concentration values

    References
    ----------
    Dutton and Maccio, "Cold dark matter haloes in the Planck era: evolution
    of structural parameters for Einasto and NFW profiles", 2014, arxiv:1402.7073
    """
    from halotools.empirical_models import NFWProfile

    if not isinstance(mass, da.Array):
        mass = da.from_array(mass, chunks=100000)

    # initialize the model
    kws = {'cosmology':cosmo.to_astropy(), 'conc_mass_model':'dutton_maccio14', 'mdef':mdef, 'redshift':redshift}
    model = NFWProfile(**kws)

    return mass.map_blocks(lambda mass: model.conc_NFWmodel(prim_haloprop=mass), dtype=mass.dtype)

def HaloRadius(mass, cosmo, redshift, mdef='vir'):
    r"""
    Return halo radius from halo mass, based on the specified mass definition.
    This is independent of halo profile, and simply returns

    .. math::

        R = \left [ 3 M /(4\pi\Delta) \right]^{1/3}

    where :math:`\Delta` is the density threshold, which depends on cosmology,
    redshift, and mass definition

    .. note::
        The units of the input mass are assumed to be :math:`M_{\odot}/h`

    Parameters
    ----------
    mass : array_like
        either a numpy or dask array specifying the halo mass; units
        assumed to be :math:`M_{\odot}/h`
    cosmo : :class:`~nbodykit.cosmology.cosmology.Cosmology`
        the cosmology instance
    redshift : float
        compute the density threshold which determines the R(M) relation
        at this redshift
    mdef : str, optional
        string specifying the halo mass definition to use; should be
        'vir' or 'XXXc' or 'XXXm' where 'XXX' is an int specifying the
        overdensity

    Returns
    -------
    radius : :class:`dask.array.Array`
        a dask array holding the halo radius
    """
    from halotools.empirical_models import halo_mass_to_halo_radius

    if not isinstance(mass, da.Array):
        mass = da.from_array(mass, chunks=100000)

    kws = {'cosmology':cosmo.to_astropy(), 'mdef':mdef, 'redshift':redshift}
    return mass.map_blocks(lambda mass: halo_mass_to_halo_radius(mass=mass, **kws), dtype=mass.dtype)

# deprecated functions
vstack = deprecate("nbodykit.transform.vstack", StackColumns, "nbodykit.transform.StackColumns")
concatenate = deprecate("nbodykit.transform.concatenate", ConcatenateSources, "nbodykit.transform.ConcatenateSources")
SkyToCartesion = deprecate("nbodykit.transform.SkyToCartesion", SkyToCartesian, "nbodykit.transform.SkyToCartesian")



from nbodykit.base.mesh import MeshSource
from nbodykit.base.catalog import CatalogSource, CatalogSourceBase
from nbodykit import _global_options
import numpy
import logging
import warnings

# for converting from particle to mesh
from pmesh import window
from pmesh.pm import RealField, ComplexField

class CatalogMesh(CatalogSource, MeshSource):
    """
    A view of a CatalogSource object which knows how to create a MeshSource
    object from itself.

    The original CatalogSource object is stored as the :attr:`base` attribute.

    Parameters
    ----------
    source : CatalogSource
        the input catalog that we are viewing as a mesh
    BoxSize :
        the size of the box
    Nmesh : int, 3-vector
        the number of cells per mesh side
    dtype : str
        the data type of the values stored on mesh
    weight : str
        column in ``source`` that specifies the weight value for each
        particle in the ``source`` to use when gridding
    value : str
        column in ``source`` that specifies the field value for each particle;
        the mesh stores a weighted average of this column
    selection : str
        column in ``source`` that selects the subset of particles to grid
        to the mesh
    position : str, optional
        column in ``source`` specifying the position coordinates; default
        is ``Position``
    interlaced : bool, optional
        use the interlacing technique of Sefusatti et al. 2015 to reduce
        the effects of aliasing on Fourier space quantities computed
        from the mesh
    compensated : bool, optional
        whether to correct for the window introduced by the grid
        interpolation scheme
    window : str, optional
        the string specifying which window interpolation scheme to use;
        see ``pmesh.window.methods``
    """
    logger = logging.getLogger('CatalogMesh')

    def __repr__(self):
        if isinstance(self.base, CatalogMesh):
            return repr(self.base)
        else:
            return "(%s as CatalogMesh)" % repr(self.base)

    def __new__(cls, source, BoxSize, Nmesh, dtype, weight,
                    value, selection, position='Position', interlaced=False,
                    compensated=False, window='cic', **kwargs):

        # source here must be a CatalogSource
        assert isinstance(source, CatalogSourceBase)

        # new, empty CatalogSource
        obj = CatalogSourceBase.__new__(cls, source.comm, source.use_cache)

        # copy over size from the CatalogSource
        obj._size = source.size
        obj._csize = source.csize

        # copy over the necessary meta-data to attrs
        obj.attrs['BoxSize'] = BoxSize
        obj.attrs['Nmesh'] = Nmesh
        obj.attrs['interlaced'] = interlaced
        obj.attrs['compensated'] = compensated
        obj.attrs['window'] = window

        # copy meta-data from source too
        obj.attrs.update(source.attrs)

        # store others as straight attributes
        obj.dtype = dtype
        obj.weight = weight
        obj.value = value
        obj.selection = selection
        obj.position = position

        # add in the Mesh Source attributes
        MeshSource.__init__(obj, obj.comm, Nmesh, BoxSize, dtype)

        # finally set the base as the input CatalogSource
        # NOTE: set this AFTER MeshSource.__init__()
        obj.base = source

        return obj

    def gslice(self, start, stop, end=1, redistribute=True):
        """
        Execute a global slice of a CatalogMesh.

        .. note::
            After the global slice is performed, the data is scattered
            evenly across all ranks.

        As CatalogMesh objects are views of a CatalogSource, this simply
        globally slices the underlying CatalogSource.

        Parameters
        ----------
        start : int
            the start index of the global slice
        stop : int
            the stop index of the global slice
        step : int, optional
            the default step size of the global size
        redistribute : bool, optional
            if ``True``, evenly re-distribute the sliced data across all
            ranks, otherwise just return any local data part of the global
            slice
        """
        # sort the base object
        newbase = self.base.gslice(start, stop, end=end, redistribute=redistribute)

        # view this base class as a CatalogMesh (with default CatalogMesh parameters)
        toret = newbase.view(self.__class__)

        # attach the meta-data from self to returned sliced CatalogMesh
        return toret.__finalize__(self)

    def sort(self, keys, reverse=False, usecols=None):
        """
        Sort the CatalogMesh object globally across all MPI ranks
        in ascending order by the input keys.

        Sort columns must be floating or integer type.

        As CatalogMesh objects are views of a CatalogSource, this simply
        sorts the underlying CatalogSource.

        Parameters
        ----------
        *keys :
            the names of columns to sort by. If multiple columns are provided,
            the data is sorted consecutively in the order provided
        reverse : bool, optional
            if ``True``, perform descending sort operations
        usecols : list, optional
            the name of the columns to include in the returned CatalogSource
        """
        # sort the base object
        newbase = self.base.sort(keys, reverse=reverse, usecols=usecols)

        # view this base class as a CatalogMesh (with default CatalogMesh parameters)
        toret = newbase.view(self.__class__)

        # attach the meta-data from self to returned sliced CatalogMesh
        return toret.__finalize__(self)

    def __slice__(self, index):
        """
        Return a slice of a CatalogMesh object.

        This slices the CatalogSource object stored as the :attr:`base`
        attribute, and then views that sliced object as a CatalogMesh.

        Parameters
        ----------
        index : array_like
            either a dask or numpy boolean array; this determines which
            rows are included in the returned object

        Returns
        -------
        subset
            the particle source with the same meta-data as ``self``, and
            with the sliced data arrays
        """
        # this slice of the CatalogSource will be the base of the mesh
        base = super(CatalogMesh, self).__slice__(index)

        # view this base class as a CatalogMesh (with default CatalogMesh parameters)
        toret = base.view(self.__class__)

        # attach the meta-data from self to returned sliced CatalogMesh
        return toret.__finalize__(self)

    def copy(self):
        """
        Return a shallow copy of ``self``.

        .. note::
            No copy of data is made.

        Returns
        -------
        CatalogMesh :
            a new CatalogMesh that holds all of the data columns of ``self``
        """
        # copy the base and view it as a CatalogMesh
        toret = self.base.copy().view(self.__class__)

        # attach the meta-data from self to returned sliced CatalogMesh
        return toret.__finalize__(self)

    def __finalize__(self, other):
        """
        Finalize the creation of a CatalogMesh object by copying over
        attributes from a second CatalogMesh.

        This also copies over the relevant MeshSource attributes via a
        call to :func:`MeshSource.__finalize__`.

        Parameters
        ----------
        obj : CatalogMesh
            the second CatalogMesh to copy over attributes from
        """
        if isinstance(other, CatalogSourceBase):
            self = CatalogSourceBase.__finalize__(self, other)

        if isinstance(other, MeshSource):
            self = MeshSource.__finalize__(self, other)

        return self

    @property
    def interlaced(self):
        """
        Whether to use interlacing when interpolating the density field.
        See :ref:`the documentation <interlacing>` for further details.

        See also: Section 3.1 of
        `Sefusatti et al. 2015 <https://arxiv.org/abs/1512.07295>`_
        """
        return self.attrs['interlaced']

    @interlaced.setter
    def interlaced(self, interlaced):
        self.attrs['interlaced'] = interlaced

    @property
    def window(self):
        """
        String specifying the name of the interpolation kernel when
        gridding the density field.

        See :ref:`the documentation <window-kernel>` for further details.

        .. note::
            Valid values must be in :attr:`pmesh.window.methods`
        """
        return self.attrs['window']

    @window.setter
    def window(self, value):
        assert value in window.methods
        self.attrs['window'] = value.lower() # lower to compare with compensation

    @property
    def compensated(self):
        """
        Boolean flag to indicate whether to correct for the windowing
        kernel introduced when interpolating the discrete particles to
        a continuous field.

        See :ref:`the documentation <compensation>` for further details.
        """
        return self.attrs['compensated']

    @compensated.setter
    def compensated(self, value):
        self.attrs['compensated'] = value

    def to_real_field(self, out=None, normalize=True):
        """
        Paint the density field, by interpolating the position column
        on to the mesh.

        This computes the following meta-data attributes in the process of
        painting, returned in the :attr:`attrs` attributes of the returned
        RealField object:

        - N : int
            the (unweighted) total number of objects painted to the mesh
        - W : float
            the weighted number of total objects, equal to the collective
            sum of the 'weight' column
        - shotnoise : float
            the Poisson shot noise, equal to the volume divided by ``N``
        - num_per_cell : float
            the mean number of weighted objects per cell

        .. note::

            The density field on the mesh is normalized as :math:`1+\delta`,
            such that the collective mean of the field is unity.

        See the :ref:`documentation <painting-mesh>` on painting for more
        details on painting catalogs to a mesh.

        Returns
        -------
        real : :class:`pmesh.pm.RealField`
            the painted real field; this has a ``attrs`` dict storing meta-data
        """
        # check for 'Position' column
        if self.position not in self:
            msg = "in order to paint a CatalogSource to a RealField, add a "
            msg += "column named '%s', representing the particle positions" %self.position
            raise ValueError(msg)

        pm = self.pm
        Nlocal = 0 # (unweighted) number of particles read on local rank
        Wlocal = 0 # (weighted) number of particles read on local rank

        # the paint brush window
        paintbrush = window.methods[self.window]

        # initialize the RealField to return
        if out is not None:
            assert isinstance(out, RealField), "output of to_real_field must be a RealField"
            numpy.testing.assert_array_equal(out.pm.Nmesh, pm.Nmesh)
            toret = out
        else:
            toret = RealField(pm)
            toret[:] = 0

        # for interlacing, we need two empty meshes if out was provided
        # since out may have non-zero elements, messing up our interlacing sum
        if self.interlaced:

            real1 = RealField(pm)
            real1[:] = 0

            # the second, shifted mesh (always needed)
            real2 = RealField(pm)
            real2[:] = 0

        # read the necessary data (as dask arrays)
        columns = [self.position, self.weight, self.value, self.selection]

        Position, Weight, Value, Selection = self.read(columns)

        # ensure the slices are synced, since decomposition is collective
        Nlocalmax = max(pm.comm.allgather(len(Position)))

        # paint data in chunks on each rank;
        # we do this by chunk 8 million is pretty big anyways.
        chunksize = _global_options['paint_chunk_size']
        for i in range(0, Nlocalmax, chunksize):
            s = slice(i, i + chunksize)

            if len(Position) != 0:

                # selection has to be computed many times when data is `large`.
                sel = self.base.compute(Selection[s])

                # be sure to use the source to compute
                position, weight, value = \
                    self.base.compute(Position[s], Weight[s], Value[s])

                # FIXME: investigate if move selection before compute
                # speeds up IO.
                position = position[sel]
                weight = weight[sel]
                value = value[sel]
            else:
                # workaround a potential dask issue on empty dask arrays
                position = numpy.empty((0, 3), dtype=Position.dtype)
                weight = None
                value = None
                selection = None

            if weight is None:
                weight = numpy.ones(len(position))

            if value is None:
                value = numpy.ones(len(position))

            # track total (selected) number and sum of weights
            Nlocal += len(position)
            Wlocal += weight.sum()

            # no interlacing
            if not self.interlaced:
                lay = pm.decompose(position, smoothing=0.5 * paintbrush.support)
                p = lay.exchange(position)
                w = lay.exchange(weight)
                v = lay.exchange(value)
                pm.paint(p, mass=w * v, resampler=paintbrush, hold=True, out=toret)

            # interlacing: use 2 meshes separated by 1/2 cell size
            else:
                lay = pm.decompose(position, smoothing=1.0 * paintbrush.support)
                p = lay.exchange(position)
                w = lay.exchange(weight)
                v = lay.exchange(value)

                H = pm.BoxSize / pm.Nmesh

                # in mesh units
                shifted = pm.affine.shift(0.5)

                # paint to two shifted meshes
                pm.paint(p, mass=w * v, resampler=paintbrush, hold=True, out=real1)
                pm.paint(p, mass=w * v, resampler=paintbrush, transform=shifted, hold=True, out=real2)

            Nglobal = pm.comm.allreduce(Nlocal)

            if pm.comm.rank == 0:
                self.logger.info("painted %d out of %d objects to mesh" 
                    % (Nglobal, self.base.csize))

        # now the loop over particles is done

        if not self.interlaced:
            # nothing to do, toret is already filled.
            pass
        else:
            # compose the two interlaced fields into the final result.
            c1 = real1.r2c()
            c2 = real2.r2c()

            # and then combine
            for k, s1, s2 in zip(c1.slabs.x, c1.slabs, c2.slabs):
                kH = sum(k[i] * H[i] for i in range(3))
                s1[...] = s1[...] * 0.5 + s2[...] * 0.5 * numpy.exp(0.5 * 1j * kH)

            # FFT back to real-space
            # NOTE: cannot use "toret" here in case user supplied "out"
            c1.c2r(real1)

            # need to add to the returned mesh if user supplied "out"
            toret[:] += real1[:]


        # unweighted number of objects
        N = pm.comm.allreduce(Nlocal)

        # weighted number of objects
        W = pm.comm.allreduce(Wlocal)

        # weighted number density (objs/cell)
        nbar = 1. * W / numpy.prod(pm.Nmesh)

        # make sure we painted something or nbar is nan; in which case
        # we set the density to uniform everywhere.
        if N == 0:
            warnings.warn(("trying to paint particle source to mesh, "
                           "but no particles were found!"),
                            RuntimeWarning
                        )

        # shot noise is volume / un-weighted number
        shotnoise = numpy.prod(pm.BoxSize) / N

        # save some meta-data
        toret.attrs = {}
        toret.attrs['shotnoise'] = shotnoise
        toret.attrs['N'] = N
        toret.attrs['W'] = W
        toret.attrs['num_per_cell'] = nbar

        csum = toret.csum()
        if pm.comm.rank == 0:
            self.logger.info("painted %d out of %d objects to mesh" %(N,self.base.csize))
            self.logger.info("mean particles per cell is %g", nbar)
            self.logger.info("sum is %g ", csum)
            self.logger.info("normalized the convention to 1 + delta")

        if normalize:
            if nbar > 0:
                toret[...] /= nbar
            else:
                toret[...] = 1

        return toret

    @property
    def actions(self):
        """
        The actions to apply to the interpolated density field, optionally
        included the compensation correction.
        """
        actions = MeshSource.actions.fget(self)
        if self.compensated:
            actions = self._get_compensation() + actions
        return actions

    def _get_compensation(self):
        """
        Return the compensation function, which corrects for the
        windowing kernel.

        The compensation function is computed as:

        - if ``interlaced = True``:
          - :func:`CompensateCIC` if using CIC window
          - :func:`CompensateTSC` if using TSC window
        - if ``interlaced = False``:
          - :func:`CompensateCICAliasing` if using CIC window
          - :func:`CompensateTSCAliasing` if using TSC window
        """
        if self.interlaced:
            d = {'cic' : self.CompensateCIC,
                 'tsc' : self.CompensateTSC}
        else:
            d = {'cic' : self.CompensateCICAliasing,
                 'tsc' : self.CompensateTSCAliasing}

        if not self.window in d:
            raise ValueError("compensation for window %s is not defined" % self.window)

        filter = d[self.window]

        return [('complex', filter, "circular")]

    @staticmethod
    def CompensateTSC(w, v):
        """
        Return the Fourier-space kernel that accounts for the convolution of
        the gridded field with the TSC window function in configuration space.

        .. note::
            see equation 18 (with p=3) of
            `Jing et al 2005 <https://arxiv.org/abs/astro-ph/0409240>`_

        Parameters
        ----------
        w : list of arrays
            the list of "circular" coordinate arrays, ranging from
            :math:`[-\pi, \pi)`.
        v : array_like
            the field array
        """
        for i in range(3):
            wi = w[i]
            tmp = (numpy.sinc(0.5 * wi / numpy.pi) ) ** 3
            v = v / tmp
        return v

    @staticmethod
    def CompensateCIC(w, v):
        """
        Return the Fourier-space kernel that accounts for the convolution of
        the gridded field with the CIC window function in configuration space

        .. note::
            see equation 18 (with p=2) of
            `Jing et al 2005 <https://arxiv.org/abs/astro-ph/0409240>`_

        Parameters
        ----------
        w : list of arrays
            the list of "circular" coordinate arrays, ranging from
            :math:`[-\pi, \pi)`.
        v : array_like
            the field array
        """
        for i in range(3):
            wi = w[i]
            tmp = (numpy.sinc(0.5 * wi / numpy.pi) ) ** 2
            tmp[wi == 0.] = 1.
            v = v / tmp
        return v

    @staticmethod
    def CompensateTSCAliasing(w, v):
        """
        Return the Fourier-space kernel that accounts for the convolution of
        the gridded field with the TSC window function in configuration space,
        as well as the approximate aliasing correction

        .. note::
            see equation 20 of
            `Jing et al 2005 <https://arxiv.org/abs/astro-ph/0409240>`_

        Parameters
        ----------
        w : list of arrays
            the list of "circular" coordinate arrays, ranging from
            :math:`[-\pi, \pi)`.
        v : array_like
            the field array
        """
        for i in range(3):
            wi = w[i]
            s = numpy.sin(0.5 * wi)**2
            v = v / (1 - s + 2./15 * s**2) ** 0.5
        return v

    @staticmethod
    def CompensateCICAliasing(w, v):
        """
        Return the Fourier-space kernel that accounts for the convolution of
        the gridded field with the CIC window function in configuration space,
        as well as the approximate aliasing correction

        .. note::
            see equation 20 of
            `Jing et al 2005 <https://arxiv.org/abs/astro-ph/0409240>`_

        Parameters
        ----------
        w : list of arrays
            the list of "circular" coordinate arrays, ranging from
            :math:`[-\pi, \pi)`.
        v : array_like
            the field array
        """
        for i in range(3):
            wi = w[i]
            v = v / (1 - 2. / 3 * numpy.sin(0.5 * wi) ** 2) ** 0.5
        return v

    def save(self, output, dataset='Field', mode='real'):
        """
        Save the mesh as a :class:`~nbodykit.source.mesh.bigfile.BigFileMesh`
        on disk, either in real or complex space.

        Parameters
        ----------
        output : str
            name of the bigfile file
        dataset : str, optional
            name of the bigfile data set where the field is stored
        mode : str, optional
            real or complex; the form of the field to store
        """
        return MeshSource.save(self, output, dataset=dataset, mode=mode)



"""
The nbodykit lab, containing all of the necessary ingredients to use nbodykit.
"""

from mpi4py import MPI
import numpy

# sources
from nbodykit.source.catalog import *
from nbodykit.source.mesh import *

# algorithms
from nbodykit.algorithms import *

from nbodykit.batch import TaskManager
from nbodykit import cosmology
from nbodykit import CurrentMPIComm
from nbodykit import transform
from nbodykit import io as IO



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
import shutil
import pytest

setup_logging()

@MPITest([4])
def test_no_seed(comm):

    CurrentMPIComm.set(comm)

    redshift = 0.55
    cosmo = cosmology.Planck15
    BoxSize = 512

    # lognormal particles
    Plin = cosmology.LinearPower(cosmo, redshift=redshift, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=3e-3, BoxSize=BoxSize, Nmesh=128)

    # seed is set randomly
    assert source.attrs['seed'] is not None

    # run FOF
    r = FOF(source, linking_length=0.2, nmin=20)
    halos = r.to_halos(cosmo=cosmo, redshift=redshift, particle_mass=1e12, mdef='vir')

    # make the HOD catalog from halotools catalog
    hod = HODCatalog(halos.to_halotools())

    # seed is set randomly
    assert hod.attrs['seed'] is not None


@MPITest([4])
def test_no_galaxies(comm):

    CurrentMPIComm.set(comm)

    redshift = 0.55
    cosmo = cosmology.Planck15
    BoxSize = 512

    Plin = cosmology.LinearPower(cosmo, redshift=redshift, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=3e-3, BoxSize=BoxSize, Nmesh=128, seed=42)

    # run FOF
    r = FOF(source, linking_length=0.2, nmin=20)
    halos = r.to_halos(cosmo=cosmo, redshift=redshift, particle_mass=1e12, mdef='vir', posdef='cm')

    # no galaxies populated into halos
    with pytest.raises(ValueError):
        hod = HODCatalog(halos.to_halotools(), seed=42, logMmin=16)

@MPITest([4])
def test_no_halos(comm):

    CurrentMPIComm.set(comm)

    redshift = 0.55
    cosmo = cosmology.Planck15
    BoxSize = 512

    # really low number density
    Plin = cosmology.LinearPower(cosmo, redshift=redshift, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=3e-5, BoxSize=BoxSize, Nmesh=128, seed=42)

    # run FOF
    r = FOF(source, linking_length=0.2, nmin=20)
    halos = r.to_halos(cosmo=cosmo, redshift=redshift, particle_mass=1e12, mdef='vir', posdef='cm')

    # no halos available
    with pytest.raises(ValueError):
        hod = HODCatalog(halos.to_halotools(), seed=42)


@MPITest([4])
def test_object_columns(comm):

    CurrentMPIComm.set(comm)

    redshift = 0.55
    cosmo = cosmology.Planck15
    BoxSize = 512

    Plin = cosmology.LinearPower(cosmo, redshift=redshift, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=3e-3, BoxSize=BoxSize, Nmesh=128, seed=42)

    # run FOF
    r = FOF(source, linking_length=0.2, nmin=20)
    halos = r.to_halos(cosmo=cosmo, redshift=redshift, particle_mass=1e12, mdef='vir', posdef='cm')

    # add an object column to the input halo catalog
    cat = halos.to_halotools()
    test =  numpy.random.random(size=len(cat.halo_table)).astype('O')
    cat.halo_table['test'] = test

    # make the HOD
    with pytest.raises(TypeError):
        hod = HODCatalog(cat, seed=42)


@MPITest([4])
def test_bad_catalog(comm):

    CurrentMPIComm.set(comm)

    redshift = 0.55
    cosmo = cosmology.Planck15
    BoxSize = 512

    Plin = cosmology.LinearPower(cosmo, redshift=redshift, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=3e-3, BoxSize=BoxSize, Nmesh=128, seed=42)

    # run FOF
    r = FOF(source, linking_length=0.2, nmin=20)
    halos = r.to_halos(cosmo=cosmo, redshift=redshift, particle_mass=1e12, mdef='vir', posdef='cm')

    # test converting from astropy to nbodykit cosmology
    cat = halos.to_halotools()
    cat.cosmology = cat.cosmology.to_astropy()

    # make the HOD
    hod = HODCatalog(cat, seed=42)

    # missing required column
    del cat.halo_table['halo_id']
    with pytest.raises(ValueError):
        hod = HODCatalog(cat, seed=42)

    # missing required attribute
    del cat.cosmology
    with pytest.raises(AttributeError):
        hod = HODCatalog(cat, seed=42)


@MPITest([4])
def test_repopulate(comm):

    CurrentMPIComm.set(comm)

    redshift = 0.55
    cosmo = cosmology.Planck15
    BoxSize = 512

    Plin = cosmology.LinearPower(cosmo, redshift=redshift, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=3e-3, BoxSize=BoxSize, Nmesh=128, seed=42)

    # run FOF
    r = FOF(source, linking_length=0.2, nmin=20)
    halos = r.to_halos(cosmo=cosmo, redshift=redshift, particle_mass=1e12, mdef='vir', posdef='cm')

    # make the HOD catalog from halotools catalog
    hod = HODCatalog(halos.to_halotools(), seed=42)
    size = hod.csize

    # repopulate (with same seed --> same catalog)
    hod.repopulate(seed=42)
    assert hod.csize == size

    # repopulate with random seed --> different catalog
    hod.repopulate()
    assert hod.csize != size

    # new params, same seed
    hod.repopulate(seed=42, alpha=1.0)
    assert hod.csize != size

    # bad param name
    with pytest.raises(ValueError):
        hod.repopulate(seed=42, bad_param_name=1.0)


@MPITest([4])
def test_hod_cm(comm):

    CurrentMPIComm.set(comm)

    redshift = 0.55
    cosmo = cosmology.Planck15
    BoxSize = 512

    # lognormal particles
    Plin = cosmology.LinearPower(cosmo, redshift=redshift, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=3e-3, BoxSize=BoxSize, Nmesh=128, seed=42)

    # run FOF
    r = FOF(source, linking_length=0.2, nmin=20)
    halos = r.to_halos(cosmo=cosmo, redshift=redshift, particle_mass=1e12, mdef='vir', posdef='cm')

    # make the HOD catalog from halotools catalog
    hod = HODCatalog(halos.to_halotools(), seed=42)

    # RSD offset in 'z' direction
    hod['Position'] += hod['VelocityOffset'] * [0, 0, 1]

    # compute the power
    r = FFTPower(hod.to_mesh(Nmesh=128), mode='2d', Nmu=5, los=[0,0,1])

@MPITest([4])
def test_hod_peak(comm):

    CurrentMPIComm.set(comm)

    redshift = 0.55
    cosmo = cosmology.Planck15
    BoxSize = 512

    # lognormal particles
    Plin = cosmology.LinearPower(cosmo, redshift=redshift, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=3e-3, BoxSize=BoxSize, Nmesh=128, seed=42)

    source['Density'] = KDDensity(source).density

    # run FOF
    r = FOF(source, linking_length=0.2, nmin=20)
    halos = r.to_halos(cosmo=cosmo, redshift=redshift, particle_mass=1e12, mdef='vir', posdef='peak')

    # make the HOD catalog from halotools catalog
    hod = HODCatalog(halos.to_halotools(), seed=42)

    # RSD offset in 'z' direction
    hod['Position'] += hod['VelocityOffset'] * [0, 0, 1]

    # compute the power
    r = FFTPower(hod.to_mesh(Nmesh=128), mode='2d', Nmu=5, los=[0,0,1])

@MPITest([4])
def test_save(comm):

    CurrentMPIComm.set(comm)

    redshift = 0.55
    cosmo = cosmology.Planck15
    BoxSize = 512

    # lognormal particles
    Plin = cosmology.LinearPower(cosmo, redshift=redshift, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=3e-3, BoxSize=BoxSize, Nmesh=128, seed=42)

    # run FOF
    r = FOF(source, linking_length=0.2, nmin=20)
    halos = r.to_halos(cosmo=cosmo, redshift=redshift, particle_mass=1e12, mdef='vir')

    # make the HOD catalog from halotools catalog
    hod = HODCatalog(halos.to_halotools(), seed=42)

    # save to a tmp file
    hod.save('tmp-hod.bigfile', ['Position', 'Velocity', 'gal_type'])

    # read tmp file
    cat = BigFileCatalog('tmp-hod.bigfile', header="Header")

    try:
        # check attrs
        for name in hod.attrs:
            numpy.testing.assert_array_equal(cat.attrs[name], hod.attrs[name])

        # check same size
        assert hod.csize == cat.csize

        # check total number of satellites
        nsat1 = comm.allreduce(hod['gal_type'].sum())
        nsat2 = comm.allreduce(cat['gal_type'].sum())
        assert nsat1 == nsat2

    finally:
        comm.barrier()
        if comm.rank == 0:
            shutil.rmtree('tmp-hod.bigfile')



from .fftpower import FFTPower, ProjectedFFTPower
from .fftcorr import FFTCorr
from .kdtree import KDDensity
from .fof import FOF
from .convpower import ConvolvedFFTPower
from .zhist import RedshiftHistogram
from .fibercollisions import FiberCollisions
from .threeptcf import Multipoles3PCF
from .sim_paircount import SimulationBoxPairCount
from .survey_paircount import SurveyDataPairCount, AngularPairCount
from .cgm import CylindricalGroups

__all__ = ['FFTPower', 'ProjectedFFTPower',
           'FFTCorr',
           'KDDensity',
           'FOF',
           'ConvolvedFFTPower',
           'RedshiftHistogram',
           'FiberCollisions',
           'Multipoles3PCF',
           'SimulationBoxPairCount',
           'SurveyDataPairCount', 'AngularPairCount',
           'CylindricalGroups'
          ]



import numpy
import logging
from six import string_types

from nbodykit import CurrentMPIComm
from nbodykit.binned_statistic import BinnedStatistic

class MissingCorrfuncError(Exception):
    def __init__(self):
        msg = "use either ``conda install -c bccp corrfunc`` "
        msg += "or ``pip install git+git://github.com/nickhand/Corrfunc``"
        self.args = (msg,)

class CorrfuncResult(object):
    """
    A class used internally holding the array like result of a
    pair counting algorithm from :mod:`Corrfunc`.

    This class is useful for summing pair count results, accounting for
    columns that are pair-weighted.

    Parameters
    ----------
    data : numpy.ndarray
        the numpy structured array result from Corrfunc
    """
    valid = ['weightavg', 'npairs', 'savg', 'ravg', 'thetaavg']

    def __init__(self, data):

        # copy over the valid colums from the input result
        dtype = [(col, data.dtype[col]) for col in data.dtype.names if col in self.valid]
        self.data = numpy.zeros(data.shape, dtype=dtype)
        self.columns = self.data.dtype.names
        for col in self.columns:
            self.data[col] = data[col]

    def __getitem__(self, col):
        return self.data[col]

    @property
    def shape(self):
        return self.data.shape

    @property
    def dtype(self):
        return self.data.dtype

    def reshape(self, *args, **kwargs):
        self.data = self.data.reshape(*args, **kwargs)
        return self

    def __radd__(self, other):
        return self + other

    def __add__(self, other):

        # all columns are pair-weighted except for "npairs"
        data = numpy.empty_like(self.data)
        for col in self.columns:
            if col == 'npairs': continue
            data[col] = self[col]*self['npairs'] + other[col]*other['npairs']

        # just sum up "npairs"
        data['npairs'] = self['npairs'] + other['npairs']

        # normalize by total "npairs"
        idx = data['npairs'] > 0.
        for col in self.columns:
            if col == 'npairs': continue
            data[col][idx] /= data['npairs'][idx]

        return CorrfuncResult(data)

class PairCountBase(object):
    """
    An abstract base class for pair counting algorithms, either for a
    simulation box or survey data.

    Users should use one of the subclasses of this class.

    Parameters
    ----------
    first : CatalogSource
        the first source of particles
    second : CatalogSource, optional
        the second source of particles to cross-correlate
    """
    def __init__(self, first, second, show_progress=True):

        self.first = first
        self.second = second
        self.comm = first.comm
        self.attrs = {}
        self.attrs['show_progress'] = show_progress

    def __getstate__(self):
        return {'result':self.result.data, 'attrs':self.attrs}

    def __setstate__(self, state):
        self.__dict__.update(state)
        redges = self.attrs['redges']
        if self.attrs['mode'] == '1d':
            self.result = BinnedStatistic(['r'], [redges], self.result, fields_to_sum=['npairs'])
        else:
            muedges = numpy.linspace(0, 1., self.attrs['Nmu']+1)
            self.result = BinnedStatistic(['r', 'mu'], [redges, muedges], self.result, fields_to_sum=['npairs'])

    def _log(self, N1, N2, pos1, pos2):
        """
        Internal function to log the distribution of particles to correlate
        across ranks.

        Parameters
        ----------
        pos1 : array_like
            the first set of particle positions this rank is correlating
        pos2 : array_like
            the second set of particle positions this rank is correlating
        """
        # global counts
        if self.comm.rank == 0:
            self.logger.info('correlating %d x %d objects in total' %(N1, N2))

        # sizes for all ranks
        sizes1 = self.comm.gather(len(pos1), root=0)
        sizes2 = self.comm.gather(len(pos2), root=0)

        # rank 0 logs
        if self.comm.rank == 0:
            args = (numpy.median(sizes1), numpy.median(sizes2))
            self.logger.info("correlating A x B = %d x %d objects (median) per rank" % args)

            global_min = numpy.min(sizes1)
            self.logger.info("min A load per rank = %d" % global_min)

            global_max = numpy.max(sizes1)
            self.logger.info("max A load per rank = %d" % global_max)

            args = (N1//self.comm.size, N2)
            self.logger.info("(even distribution would result in %d x %d)" % args)

    def _decompose(self):
        """
        An internal function to perform a domain decomposition on the two
        sources we wish to correlate, returning the position/weights for each
        object in the correlating pair.

        The particles are assumed to be in a cubical box, and no load
        balancing is required.

        The implementation follows:

        1. Decompose the first source such that the objects are spatially
           tight on a given rank.
        2. Decompose the second source, ensuring a given rank holds all
           particles within the desired maximum separation.

        Returns
        -------
        (pos1, w1), (pos2, w2) : array_like
            the (decomposed) set of positions and weights to correlate
        """
        from pmesh.domain import GridND
        from nbodykit.utils import split_size_3d

        # determine processor division for domain decomposition
        np = split_size_3d(self.comm.size)
        if self.comm.rank == 0:
            self.logger.info("using cpu grid decomposition: %s" %str(np))

        # get the (periodic-enforced) position for first
        pos1 = self.first['Position']
        if self.attrs['periodic']:
            pos1 %= self.attrs['BoxSize']
        pos1, w1 = self.first.compute(pos1, self.first[self.attrs['weight']])
        N1 = self.comm.allreduce(len(pos1))

        # get the (periodic-enforced) position for second
        if self.second is not None:
            pos2 = self.second['Position']
            if self.attrs['periodic']:
                pos2 %= self.attrs['BoxSize']
            pos2, w2 = self.second.compute(pos2, self.second[self.attrs['weight']])
            N2 = self.comm.allreduce(len(pos2))
        else:
            pos2 = pos1
            w2 = w1
            N2 = N1

        # domain decomposition
        grid = [
            numpy.linspace(0, self.attrs['BoxSize'][0], np[0] + 1, endpoint=True),
            numpy.linspace(0, self.attrs['BoxSize'][1], np[1] + 1, endpoint=True),
            numpy.linspace(0, self.attrs['BoxSize'][2], np[2] + 1, endpoint=True),
        ]
        domain = GridND(grid, comm=self.comm)

        # exchange first particles
        layout = domain.decompose(pos1, smoothing=0)
        pos1 = layout.exchange(pos1)
        w1 = layout.exchange(w1)

        # exchange second particles
        rmax = numpy.max(self.attrs['redges'])
        if rmax > self.attrs['BoxSize'].max() * 0.25:
            pos2 = numpy.concatenate(self.comm.allgather(pos2), axis=0)
            w2   = numpy.concatenate(self.comm.allgather(w2), axis=0)
        else:
            layout  = domain.decompose(pos2, smoothing=rmax)
            pos2 = layout.exchange(pos2)
            w2   = layout.exchange(w2)

        # log the sizes of the trees
        self._log(N1, N2, pos1, pos2)

        return (pos1, w1), (pos2, w2)

    def _run(self, func, kwargs, loads, callback=None):
        """
        Internal function that calls ``func`` in ``N`` iterations, optionally
        calling ``callback`` before each iteration.

        This allows the Corrfunc function ``func`` to be called in chunks,
        giving the user a progress report after each iteration.

        Parameters
        ----------
        func : callable
            the Corrfunc callable -- ``kwargs`` will be passed ot this function
        kwargs : dict
            the dictionary of arguments to pass to ``func``
        loads : list of int
            the list of loads for every rank; this corresponds to the number
            of particles in the in ``A`` if we are correlating ``A`` x ``B``
        callback : callable, optional
            a callable takings ``kwargs`` as its first argument and a slice
            object as its second argument; this will be called first during
            each iteration

        Returns
        -------
        pc : CorrfuncResult
            the total pair count result, summed over all iterations run
        """
        # the rank with the largest load
        largest_load = numpy.argmax(loads)

        # do the pair counting
        def run(chunk):
            if callback is not None:
                callback(kwargs, chunk)
            return self._run_corrfunc(func, **kwargs)

        # log the function start
        if self.comm.rank == 0:
            name = func.__module__ + '.' + func.__name__
            self.logger.info("calling function '%s'" % name)

        # number of iterations
        N = 10 if self.attrs['show_progress'] else 1

        # run in chunks
        pc = None
        chunks = numpy.array_split(range(loads[self.comm.rank]), N, axis=0)
        for i, chunk in enumerate(chunks):
            this_pc = run(chunk)
            if self.comm.rank == largest_load and self.attrs['show_progress']:
                self.logger.info("%d%% done" % (N*(i+1)))

            # sum up the results
            pc = this_pc if pc is None else pc + this_pc

        return pc

    def _run_corrfunc(self, func, **kws):
        """
        Run the wrapped Corrfunc function ``func``, passing in the keywords
        specified by ``kws``.

        This hides all output from the Corrfunc function (stdout, stderr,
        and C-level output), unless an exception occurs.
        """
        from nbodykit.utils import captured_output

        try:
            # record progress capture output for everything but root
            with captured_output(self.comm, root=None) as (out, err):
                result = func(**kws)

        except Exception as e:
            # get the values Corrfunc logged to stdout and stderr
            stdout = out.getvalue(); stderr = err.getvalue()

            # log all of the output in a new exception
            name = func.__module__ + '.' + func.__name__
            msg = "calling the function '%s' failed:\n" % name
            msg += "exception: %s\n" % str(e)
            msg += "stdout: %s\n" % stdout
            msg += "stderr: %s" % stderr
            raise RuntimeError(msg)

        return CorrfuncResult(result)

    def save(self, output):
        """
        Save result as a JSON file with name ``output``
        """
        import json
        from nbodykit.utils import JSONEncoder

        # only the master rank writes
        if self.comm.rank == 0:
            self.logger.info('measurement done; saving result to %s' % output)

            with open(output, 'w') as ff:
                json.dump(self.__getstate__(), ff, cls=JSONEncoder)

    @classmethod
    @CurrentMPIComm.enable
    def load(cls, output, comm=None):
        """
        Load a result has been saved to disk with :func:`save`.
        """
        import json
        from nbodykit.utils import JSONDecoder
        if comm.rank == 0:
            with open(output, 'r') as ff:
                state = json.load(ff, cls=JSONDecoder)
        else:
            state = None
        state = comm.bcast(state)
        self = object.__new__(cls)
        self.__setstate__(state)
        self.comm = comm
        return self


class SimulationBoxPairCount(PairCountBase):
    """
    Count (weighted) pairs of objects in a simulation box using the
    :mod:`Corrfunc` package.

    This uses the :func:`Corrfunc.theory.DD.DD` and
    :func:`Corrfunc.theory.DDsmu.DDsmu` functions to count pairs.

    Results are computed when the object is inititalized. See the documenation
    of :func:`~SimulationBoxPairCount.run` for the attributes storing the
    results.

    .. note::

        The algorithm expects the positions of particles in a simulation box to
        be the Cartesian ``x``, ``y``, and ``z`` vectors. To compute
        pair counts on survey data, using right ascension, declination, and
        redshift, see :class:`SurveyDataPairCount`.

    Parameters
    ----------
    mode : {'1d', '2d'}
        compute pair counts as a function of ``r`` and ``mu`` or just ``r``
    first : CatalogSource
        the first source of particles, providing the 'Position' column
    redges : array_like
        the radius bin edges; length of nbins+1
    BoxSize : float, 3-vector, optional
        the size of the box; if 'BoxSize' is not provided in the source
        'attrs', it must be provided here
    second : CatalogSource, optional
        the second source of particles to cross-correlate
    Nmu : int, optional
        the number of ``mu`` bins, ranging from 0 to 1
    los : {'x', 'y', 'z'}, int, optional
        the axis of the simulation box to treat as the line-of-sight direction;
        this can be provided as string identifying one of 'x', 'y', 'z' or
        the equivalent integer number of the axis
    periodic : bool, optional
        whether to use periodic boundary conditions
    weight : str, optional
        the name of the column in the source specifying the particle weights
    show_progress : bool, optional
        if ``True``, perform the pair counting calculation in 10 iterations,
        logging the progress after each iteration; this is useful for
        understanding the scaling of the code
    **config : key/value pairs
        additional keywords to pass to the :mod:`Corrfunc` function
    """
    logger = logging.getLogger('SimulationBoxPairCount')

    def __init__(self, mode, first, redges, BoxSize=None, periodic=True,
                    second=None, Nmu=5, los='z', weight='Weight',
                    show_progress=True, **config):

        # check input 'mode'
        assert mode in ['1d', '2d'], "PairCount mode must be '1d' or '2d'"

        # check rmin
        if numpy.min(redges) <= 0.:
            raise ValueError(("the lower edge of the 1st separation bin must "
                              "greater than zero (no self-pairs)"))

        # check input 'los'
        if isinstance(los, string_types):
            assert los in 'xyz', "``los`` should be one of 'x', 'y', 'z'"
            los = 'xyz'.index(los)
        elif isinstance(los, int):
            if los < 0: los += 3
            assert los in [0,1,2], "``los`` should be one of 0, 1, 2"
        else:
            raise ValueError("``los`` should be either ['x', 'y', 'z'] or [0,1,2]")

        # verify the input sources
        BoxSize = verify_input_sources(first, second, BoxSize, ['Position', weight])

        # init the base class
        PairCountBase.__init__(self, first, second, show_progress)

        # save the meta-data
        self.attrs['mode'] = mode
        self.attrs['redges'] = redges
        self.attrs['Nmu'] = Nmu
        self.attrs['BoxSize'] = BoxSize
        self.attrs['periodic'] = periodic
        self.attrs['weight'] = weight
        self.attrs['config'] = config
        self.attrs['los'] = los

        # test Rmax for PBC
        if periodic and numpy.amax(redges) > 0.5*self.attrs['BoxSize'].min():
            raise ValueError(("periodic pair counts cannot be computed for "
                              "Rmax > 0.5 * BoxSize"))

        # run the algorithm
        self.run()

    def run(self):
        """
        Calculate the 3D pair-counts in a simulation box as a function
        of separation ``r`` or separation and angle to line-of-sight
        (``r``, ``mu``). This adds the following attributes to the class:

        - :attr:`SimulationBoxPairCount.result`

        Attributes
        ----------
        result : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            a BinnedStatistic object holding the pair count and correlation
            function results. The coordinate grid is either ``r`` or
            ``r`` and ``mu``. It stores the following variables:

            - ``r``: the mean separation value in the bin
            - ``xi``: the mean correlation function value in the bin, computed as
              :math:`DD/RR - 1`, where :math:`RR` is the number of random pairs
              in the bin
            - ``npairs``: the number of pairs in the bin
            - ``weightavg``: the average weight value in the bin; each pair
              contributes the product of the individual weight values
        """
        try:
            from Corrfunc.theory import DD, DDsmu
        except ImportError:
            raise MissingCorrfuncError()

        # some setup
        redges = self.attrs['redges']
        boxsize = self.attrs['BoxSize']
        L = None

        # determine axes perp and paralell to
        los = self.attrs['los']
        perp_axes = [0,1,2]
        perp_axes.remove(los)

        # cubic box required by Corrunc
        if self.attrs['periodic']:
            if numpy.all(boxsize == boxsize[0]):
                L = boxsize[0]
            else:
                raise ValueError(("``Corrfunc`` does not currently support  "
                                  "periodic wrapping with non-cubic boxes"))

        # domain decompose the data
        (pos1, w1), (pos2, w2) = self._decompose()

        # total sizes for first and second
        N1 = self.comm.allreduce(len(pos1))
        N2 = self.comm.allreduce(len(pos2))

        # initialize the keywords to pass to Corrfunc
        kws = {}
        kws['autocorr'] = 0
        kws['nthreads'] = 1
        kws['binfile'] = redges
        kws['X2'] = pos2[:,perp_axes[0]]
        kws['Y2'] = pos2[:,perp_axes[1]]
        kws['Z2'] = pos2[:,los]
        kws['weights2'] = w2.astype(pos2.dtype)
        kws['weight_type'] = 'pair_product'
        kws['periodic'] = self.attrs['periodic']
        if L is not None: kws['boxsize'] = L

        # 1D calculation
        if self.attrs['mode'] == '1d':
            kws['output_ravg'] = True
            func = DD
            rcol = 'ravg'
        # 2D calculation
        else:
            kws['output_savg'] = True
            kws['mu_max'] = 1.0
            kws['nmu_bins'] = self.attrs['Nmu']
            func = DDsmu
            rcol = 'savg'

        # update with user config
        kws.update(self.attrs['config'])

        def callback(kws, chunk):
            kws['X1'] = pos1[chunk][:,perp_axes[0]]
            kws['Y1'] = pos1[chunk][:,perp_axes[1]]
            kws['Z1'] = pos1[chunk][:,los] # LOS defined with respect to this axis
            kws['weights1'] = w1[chunk].astype(pos1.dtype)

        # run
        sizes = self.comm.allgather(len(pos1))
        pc = self._run(func, kws, sizes, callback=callback)

        # make 2D results shape of (Nr, Nmu)
        if self.attrs['mode'] == '2d':
            pc = pc.reshape((-1, self.attrs['Nmu']))

        # sum results over all ranks
        pc = self.comm.allreduce(pc)

        # the output data
        dtype = numpy.dtype([('r', 'f8'), ('xi', 'f8'), ('npairs', 'u8'), ('weightavg', 'f8')])
        data = numpy.empty(pc.shape, dtype=dtype)

        # copy over main results
        data['r'] = pc[rcol]
        data['npairs'] = pc['npairs']
        data['weightavg'] = pc['weightavg']

        # compute the random pairs from the fractional volume
        RR = 1.*N1*N2 / boxsize.prod()
        if self.attrs['mode'] == '2d':
            dr3 = numpy.diff(redges**3)
            muedges = numpy.linspace(0, 1, self.attrs['Nmu']+1)
            dmu = numpy.diff(muedges)
            RR *= 2. / 3. * numpy.pi * dr3[:,None] * dmu[None,:]
        else:
            RR *= 4. / 3. * numpy.pi * numpy.diff(redges**3)

        # add in the correlation function value
        data['xi'] = (1. * data['npairs'] / RR) - 1.0

        # make the BinnedStatistic
        if self.attrs['mode'] == '1d':
            self.result = BinnedStatistic(['r'], [redges], data, fields_to_sum=['npairs'])
        else:
            self.result = BinnedStatistic(['r','mu'], [redges,muedges], data, fields_to_sum=['npairs'])


def verify_input_sources(first, second, BoxSize, required_columns, inspect_boxsize=True):
    """
    Verify that the input source objects have all of the required columns
    and appropriate box size attributes.
    """
    if second is None: second = first

    # check for comm mismatch
    assert second.comm is first.comm, "communicator mismatch between input sources"

    # check required columns
    for source in [first, second]:
        for col in required_columns:
            if col not in source:
                raise ValueError("the column '%s' is missing from input source; cannot do pair count" %col)

    if inspect_boxsize:
        _BoxSize = numpy.zeros(3)
        BoxSize1 = first.attrs.get('BoxSize', None)
        if BoxSize1 is not None:
            _BoxSize[:] = BoxSize1[:]
        if BoxSize is not None:
            _BoxSize[:] = BoxSize
        if (_BoxSize==0.).all():
            raise ValueError("BoxSize must be supplied in the source ``attr`` or via the ``BoxSize`` keyword")

        BoxSize2 = second.attrs.get('BoxSize', None)
        if BoxSize1 is not None and BoxSize2 is not None:
            if not numpy.array_equal(first.attrs['BoxSize'], BoxSize2):
                raise ValueError("BoxSize mismatch between pair count cross-correlation sources")
            if not numpy.array_equal(first.attrs['BoxSize'], _BoxSize):
                raise ValueError("BoxSize mismatch between sources and the pair count algorithm")

        return _BoxSize
